﻿using System;
using System.Text;

using System.Net;
using System.Net.Sockets;

namespace MultiDevice
{
    class UDP
    {
        public struct Device
        {
            public string       Host_Name;
            public IPAddress    IP;
            public IPAddress    Mask;
            public IPAddress    Gateway;
            public string       MAC;
            public byte         Blocked;
            public ushort       HTTP_Port;
            public float        Temperatur;
            public string       FirmWare;
            public char         Typ;

            public bool  Power_Measurement;
            public float VoltageRMS;
            public float CurrentRMS;
            public float LineFrequency;
            public float ActivePower;
            public float ApparentPower;
            public float ReactivePower;
            public float PowerFactor;

            public bool  Sensor_1_Ready;
            public float Sensor_1_Temperature;
            public float Sensor_1_Humidity;
            public int   Sensor_1_Brightness;
            
            public int   Array_Pos;
            public bool  Save_Power_Measurement;
            public long  TimeStamp;
            public bool  XOR_USER_Password;

        }

        public struct Device_Socket
        {
            public string   Socket_Name;
            public bool     Socket_State;
            public string   IO_Name;
            public bool     IO_Direction;
            public bool     IO_State;
        }

        public const byte DevNo = 20;
        public static Device[] Devices = new Device[DevNo];
        public static Device_Socket[] Sockets = new Device_Socket[DevNo * 8];
        public string[] buf_split;

        public bool Received = false;
        public string NoPass= "";

        public static string[] File_Time_Array = new string[DevNo];
        public static string[,] Array_Time = new string[3600, DevNo];
        public static float[,] Array_VoltageRMS = new float[3600, DevNo];
        public static float[,] Array_CurrentRMS = new float[3600, DevNo];
        public static float[,] Array_LineFrequency = new float[3600, DevNo];
        public static float[,] Array_ActivePower = new float[3600, DevNo];
        public static float[,] Array_ApparentPower = new float[3600, DevNo];
        public static float[,] Array_ReactivePower = new float[3600, DevNo];
        public static float[,] Array_PowerFactor = new float[3600, DevNo];

        public static Device[] UDP_Devices
        {
            get { return Devices; }
            set { Devices = value; }
        }

        public static Device_Socket[] UDP_Sockets
        {
            get { return Sockets; }
            set { Sockets = value; }
        }

        public async void UDPListener(int listenPort)
        {
            try
            {
                using (var udpClient = new UdpClient(listenPort))
                {
                    while (true)
                    {
                        try
                        {
                            var receivedResults = await udpClient.ReceiveAsync();
                            byte[] receivedString = receivedResults.Buffer;
                            Parse_Answer(receivedString);
                        }
                        catch { }
                    }
                }
            }
            catch { }
        }

        //Decode packet
        private void Parse_Answer(byte[] receivedString)
        {
            byte idx = 0;
            bool not_exists = true;

            //Split received message.
            buf_split = Encoding.UTF8.GetString(receivedString).Split(':');

            //Not NET-PwrCtrl
            if (buf_split[0] != "NET-PwrCtrl") return;

            //Error wrong User;Pass
            if (buf_split[2] == "NoPass")
            {
                NoPass = buf_split[1]; //NoPass=Hostname else NoPass=empty string
                Received = true;
                return;
            }

            //power measurement refresh
            //answer after "pow"
            if(buf_split.Length<12)
            {
                byte i;

                for (i = 0; i < UDP.DevNo; i++)
                    if (Devices[i].Host_Name == buf_split[1]) break;

                Read_Power(i, 3);
                return;
            }

            //Search free position
            for (byte i = 0; i < DevNo; i++)
            {
                if (Devices[i].Host_Name == null)
                {
                    idx = i;
                    break;
                }
            }

            //Host_Name exists -> put in the same cell
            for (byte i = 0; i < DevNo; i++)
            {
                if (Devices[i].Host_Name == buf_split[1])
                {
                    not_exists = false;
                    idx = i;
                    break;
                }
            }

            if (not_exists) System.Diagnostics.Debug.WriteLine(receivedString);

            // Flag set for update UI
            Received = true;

            //Parse parameter
            Devices[idx].Host_Name = buf_split[1];
            Devices[idx].IP = IPAddress.Parse(buf_split[2]);

            Devices[idx].Mask = IPAddress.Parse(buf_split[3]);
            Devices[idx].Gateway = IPAddress.Parse(buf_split[4]);
            Devices[idx].MAC = buf_split[5];

            for (int i = 0; i < 8; i++)
            {
                string[] name_split = buf_split[6 + i].Split(',');
                Sockets[i + (idx * 8)].Socket_Name = name_split[0];
                Sockets[i + (idx * 8)].Socket_State = Convert.ToBoolean(Int16.Parse(name_split[1]));
            }

            Devices[idx].Blocked = Convert.ToByte(buf_split[14]);
            Devices[idx].HTTP_Port = Convert.ToUInt16(buf_split[15]);
            Devices[idx].Temperatur = -127;

            //HOME, PRO 4.5
            if (buf_split[16].StartsWith("NET-PWRCTRL") == true)
            {
                Devices[idx].Typ = Convert.ToChar(buf_split[17]);

                Devices[idx].XOR_USER_Password = false;
                if (buf_split[18] == "xor")
                    Devices[idx].XOR_USER_Password = true;

            }
            else
            {

                if (Devices[idx].Blocked >= 248)
                    Devices[idx].Typ = 'H';
                else
                    Devices[idx].Typ = 'P';
            }

            //More than HOME/PRO
            if (buf_split.Length > 20)
            {
                //IO
                for (byte i = 0; i < 8; i++)
                {
                    string[] name_split = buf_split[16 + i].Split(',');
                    Sockets[i + (idx * 8)].IO_Name = name_split[0];
                    Sockets[i + (idx * 8)].IO_Direction = Convert.ToBoolean(Int16.Parse(name_split[1]));
                    Sockets[i + (idx * 8)].IO_State = Convert.ToBoolean(Int16.Parse(name_split[2]));
                }

                string temp = buf_split[24].Remove(buf_split[24].Length - 2);
                Devices[idx].Temperatur = Convert.ToSingle(temp.Replace(".", ","));
                Devices[idx].FirmWare = buf_split[25];

                //since FW 6.0
                if (buf_split.Length > 26)
                {
                    Devices[idx].Typ = Convert.ToChar(buf_split[26]);

                    //--------------------------------------------------------------------------------
                    //Typ { a = ADV; i = IO; h = HUT; o = ONE; f = ONE-F; H = Home; P = PRO}
                    //--------------------------------------------------------------------------------

                    //Power
                    Devices[idx].Power_Measurement = false;
                    if (buf_split[27] == "p")
                    {
                        Read_Power(idx, 28);
                        Devices[idx].Power_Measurement = true;
                    }

                    //Sensor 1
                    Devices[idx].Sensor_1_Ready = false;
                    if (buf_split[buf_split.Length-6] == "s")
                    {
                        Devices[idx].Sensor_1_Ready = true;
                        Devices[idx].Sensor_1_Temperature = Convert.ToSingle(buf_split[buf_split.Length - 5].Replace(".", ","));
                        Devices[idx].Sensor_1_Humidity = Convert.ToSingle(buf_split[buf_split.Length - 4].Replace(".", ","));
                        Devices[idx].Sensor_1_Brightness = Convert.ToUInt16(buf_split[buf_split.Length - 3].Replace(".", ","));
                    }

                    Devices[idx].XOR_USER_Password = false;
                    if (buf_split[buf_split.Length-2] == "xor")
                    {
                        Devices[idx].XOR_USER_Password = true;
                    }
                    return;
                }
            }
            System.Diagnostics.Debug.WriteLine(Devices[idx].Typ);
        }

        public void Read_Power(byte idx, int start)
        {
            Devices[idx].VoltageRMS = Convert.ToSingle(buf_split[start].Replace(".", ","));
            Devices[idx].CurrentRMS = Convert.ToSingle(buf_split[start + 1].Replace(".", ","));
            Devices[idx].LineFrequency = Convert.ToSingle(buf_split[start + 2].Replace(".", ","));
            Devices[idx].ActivePower = Convert.ToSingle(buf_split[start + 3].Replace(".", ","));
            Devices[idx].ApparentPower = Convert.ToSingle(buf_split[start + 4].Replace(".", ","));
            Devices[idx].ReactivePower = Convert.ToSingle(buf_split[start + 5].Replace(".", ","));
            Devices[idx].PowerFactor = Convert.ToSingle(buf_split[start + 6].Replace(".", ","));
        }

        public void Send_Ask(int SenderPort, string command)
        {
            UdpClient client = new UdpClient();
            IPEndPoint ip = new IPEndPoint(IPAddress.Broadcast, SenderPort);
            byte[] bytes = Encoding.ASCII.GetBytes(command);
            client.Send(bytes, bytes.Length, ip);
            client.Close();
        }

        public void Switch(int SenderPort, IPAddress IP, string command, int socket, string user, string password, string encrypt)
        {
            if (IP == null) return;

            UdpClient sw_client = new UdpClient();
            IPEndPoint ip = new IPEndPoint(IP, SenderPort);

            string user_password = user + password;
            string encr_user_password = EncryptOrDecrypt(user_password, password)+ '\0';

            string bef;
            if (encrypt == "⌘")
                bef = command + socket + encr_user_password; //for example: Sw_on1ASfdhhgfDS
            else
                bef = command + socket + user + password; //for example: Sw_on1adminanel

            byte[] bytes = Encoding.ASCII.GetBytes(bef);
            sw_client.Send(bytes, bytes.Length, ip);
            sw_client.Close();
        }

        public void Switch_all(int SenderPort, IPAddress IP, string command, byte sockets, string user, string password, string encrypt)
        {
            if (IP == null) return;

            UdpClient sw_client = new UdpClient();
            IPEndPoint ip = new IPEndPoint(IP, SenderPort);

            string user_password = user + password;
            string encr_user_password = EncryptOrDecrypt(user_password, password) + '\0';

            string bef;
            if (encrypt == "⌘")
                bef = command + " " + encr_user_password; //for example: Sw_on1ASfdhhgfDS
            else
                bef = command + " " + user + password; //for example: Sw_on1adminanel

            byte[] bytes = Encoding.ASCII.GetBytes(bef);
            bytes[2] = sockets; //state of all sockets as a byte  BitNo   76543210
                                //                                Socket  87654321
            sw_client.Send(bytes, bytes.Length, ip);
            sw_client.Close();
        }

        string EncryptOrDecrypt(string text, string key)
        {
            var result = new StringBuilder();

            for (int c = 0; c < text.Length; c++)
                result.Append((char)((uint)text[c] ^ (uint)key[c % key.Length]));

            byte[] buffer = Encoding.ASCII.GetBytes(result.ToString());

            return Convert.ToBase64String(buffer);
        }
    }
}
